#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
from abc import ABC, abstractmethod
from collections import Iterable
from typing import Union, Tuple

from notification_service.base_notification import BaseEvent, ANY_CONDITION, SenderEventCount
from notification_service.util import db
from notification_service.util.db import EventModel, ClientModel, MemberModel


class BaseEventStorage(ABC):

    @abstractmethod
    def add_event(self, event: BaseEvent, uuid: str):
        pass

    @abstractmethod
    def list_events(self,
                    key: Union[str, Tuple[str]],
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    namespace: str = None,
                    sender: str = None):
        pass

    @abstractmethod
    def count_events(self,
                     key: Union[str, Tuple[str]],
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     namespace: str = None,
                     sender: str = None):
        pass

    @abstractmethod
    def list_all_events(self, start_time: int):
        pass

    @abstractmethod
    def list_all_events_from_version(self, start_version: int, end_version: int = None):
        pass

    @abstractmethod
    def clean_up(self):
        pass

    @abstractmethod
    def get_latest_version(self, key: str, namespace: str = None):
        pass

    @abstractmethod
    def register_client(self, namespace: str = None, sender: str = None) -> int:
        pass

    @abstractmethod
    def delete_client(self, client_id):
        pass

    @abstractmethod
    def is_client_exists(self, client_id) -> bool:
        pass

    def get_event_by_uuid(self, uuid: str):
        pass


class MemoryEventStorage(BaseEventStorage):

    def __init__(self):
        self.store = []
        self.store_with_uuid = {}
        self.max_version = 0
        self.clients = []

    def add_event(self, event: BaseEvent, uuid: str):
        self.max_version += 1
        event.create_time = int(time.time() * 1000)
        event.version = self.max_version
        self.store.append(event)
        self.store_with_uuid[uuid] = event
        return event

    def list_events(self,
                    key: Union[str, Tuple[str]],
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    namespace: str = None,
                    sender: str = None):
        res = []
        key = None if key == "" else key
        version = None if version == 0 else version
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        for event in self.store:
            if key is not None and event.key not in key and ANY_CONDITION not in key:
                continue
            if version is not None and event.version <= version:
                continue
            if event_type is not None and event.event_type != event_type and event_type != ANY_CONDITION:
                continue
            if start_time is not None and event.create_time < start_time:
                continue
            if namespace is not None and namespace != ANY_CONDITION and event.namespace != namespace:
                continue
            if sender is not None and sender != ANY_CONDITION and event.sender != sender:
                continue
            res.append(event)
        return res

    def count_events(self,
                     key: Union[str, Tuple[str]],
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     namespace: str = None,
                     sender: str = None):
        key = None if key == "" else key
        version = None if version == 0 else version
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        event_counts = {}
        for event in self.store:
            if key is not None and event.key not in key and ANY_CONDITION not in key:
                continue
            if version is not None and event.version <= version:
                continue
            if event_type is not None and event.event_type != event_type and event_type != ANY_CONDITION:
                continue
            if start_time is not None and event.create_time < start_time:
                continue
            if namespace is not None and namespace != ANY_CONDITION and event.namespace != namespace:
                continue
            if sender is not None and sender != ANY_CONDITION and event.sender != sender:
                continue
            if event.sender in event_counts:
                event_counts.update({event.sender: event_counts.get(event.sender) + 1})
            else:
                event_counts.update({event.sender: 1})
        res = []
        for sender, event_count in event_counts.items():
            res.append(SenderEventCount(sender=sender, event_count=event_count))
        return res

    def list_all_events(self, start_time: int):
        res = []
        for event in self.store:
            if event.create_time >= start_time:
                res.append(event)
        return res

    def list_all_events_from_version(self, start_version: int, end_version: int = None):
        res = []
        for event in self.store:
            if 0 < end_version < event.version:
                continue
            if event.version > start_version:
                res.append(event)
        return res

    def get_latest_version(self, key: str, namespace: str = None):
        return self.max_version

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        self.clients.append((len(self.clients), namespace, sender))
        return len(self.clients) - 1

    def delete_client(self, client_id):
        self.clients.pop(client_id)

    def is_client_exists(self, client_id) -> bool:
        for item in self.clients:
            if item[0] == client_id:
                return True
        return False

    def get_event_by_uuid(self, uuid: str):
        return self.store_with_uuid.get(uuid)

    def clean_up(self):
        self.store.clear()
        self.clients.clear()
        self.store_with_uuid.clear()
        self.max_version = 0


class DbEventStorage(BaseEventStorage):

    def __init__(self, db_conn=None):
        if db_conn is not None:
            db.SQL_ALCHEMY_CONN = db_conn
        db.prepare_db()

    def add_event(self, event: BaseEvent, uuid: str):
        return EventModel.add_event(event, uuid)

    def list_events(self,
                    key: Union[str, Tuple[str]],
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    namespace: str = None,
                    sender: str = None):
        return EventModel.list_events(key, version, event_type, start_time, namespace, sender)

    def count_events(self,
                     key: Union[str, Tuple[str]],
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     namespace: str = None,
                     sender: str = None):
        return EventModel.count_events(key, version, event_type, start_time, namespace, sender)

    def list_all_events(self, start_time: int):
        return EventModel.list_all_events(start_time)

    def list_all_events_from_version(self, start_version: int, end_version: int = None):
        return EventModel.list_all_events_from_version(start_version, end_version)

    def get_latest_version(self, key: str, namespace: str = None):
        return EventModel.get_latest_version()

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        return ClientModel.register_client(namespace, sender)

    def delete_client(self, client_id):
        ClientModel.delete_client(client_id)

    def is_client_exists(self, client_id) -> bool:
        return ClientModel.is_client_exists(client_id)

    def get_event_by_uuid(self, uuid: str):
        return EventModel.get_event_by_uuid(uuid)

    def clean_up(self):
        EventModel.cleanup()
        ClientModel.cleanup()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import asyncio
import functools
import inspect
import logging
import os
import threading
import time
from concurrent import futures

import grpc
from grpc import _common, _server
from grpc._cython.cygrpc import StatusCode
from grpc._server import _serialize_response, _status, _abort, _Context, _unary_request, \
    _select_thread_pool_for_behavior, _unary_response_in_pool

from notification_service.event_storage import DbEventStorage
from notification_service.high_availability import DbHighAvailabilityStorage, SimpleNotificationServerHaManager
from notification_service.proto import notification_service_pb2_grpc
from notification_service.service import NotificationService, HighAvailableNotificationService
from notification_service.server_config import NotificationServerConfig
from notification_service.util.utils import get_ip_addr

_PORT = 50051
_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class NotificationServer(object):
    """
    Block/Async server of Notification function service.
    """

    def __init__(self, service, port=_PORT):
        self.executor = Executor(futures.ThreadPoolExecutor(max_workers=10))
        self.grpc_server = grpc.server(self.executor)
        self.service = service
        notification_service_pb2_grpc.add_NotificationServiceServicer_to_server(service,
                                                                                self.grpc_server)
        self.grpc_server.add_insecure_port('[::]:' + str(port))

    def run(self, is_block=False):
        """
        start the notification service
        :param is_block: is block mode
        :return:
        """
        self.service.start()
        self.grpc_server.start()
        logging.info('Notification server started.')
        if is_block:
            try:
                while True:
                    time.sleep(_ONE_DAY_IN_SECONDS)
            except KeyboardInterrupt:
                self.stop()
        else:
            pass

    def stop(self):
        """
        stop the notification service
        :return:
        """
        self.executor.shutdown()
        self.service.stop()
        self.grpc_server.stop(0)
        logging.info('Notification server stopped.')


class NotificationServerRunner(object):
    def __init__(self, config_file):
        if not os.path.exists(config_file):
            raise IOError('Config file {} not exist!'.format(config_file))
        self.config = NotificationServerConfig(config_file=config_file)

    def _init_server(self):
        if self.config.db_uri:
            self.storage = DbEventStorage(self.config.db_uri)
        else:
            raise Exception('Failed to start notification service without database connection info.')
        if self.config.enable_ha:
            server_uri = self.config.advertised_uri \
                if self.config.advertised_uri is not None else get_ip_addr() + ':' + str(self.config.port)
            ha_storage = DbHighAvailabilityStorage(db_conn=self.config.db_uri)
            ha_manager = SimpleNotificationServerHaManager()
            service = HighAvailableNotificationService(
                self.storage,
                ha_manager,
                server_uri,
                ha_storage,
                5000)
            self.server = NotificationServer(service=service,
                                             port=int(self.config.port))
        else:
            self.server = NotificationServer(service=NotificationService(self.storage),
                                             port=int(self.config.port))

    def start(self, is_block=False):
        self._init_server()
        self.server.run(is_block)

    def stop(self):
        self.server.stop()


def _loop(loop: asyncio.AbstractEventLoop):
    asyncio.set_event_loop(loop)
    if not loop.is_running() or loop.is_closed():
        loop.run_forever()
    pending = asyncio.all_tasks(loop=loop)
    if pending:
        loop.run_until_complete(asyncio.gather(*pending))


class Executor(futures.Executor):
    def __init__(self, thread_pool, loop=None):
        super().__init__()
        self._shutdown = False
        self._thread_pool = thread_pool
        self._loop = loop or asyncio.get_event_loop()
        if not self._loop.is_running() or self._loop.is_closed():
            self._thread = threading.Thread(target=_loop, args=(self._loop,), daemon=True)
            self._thread.start()

    def submit(self, fn, *args, **kwargs):
        if self._shutdown:
            raise RuntimeError('Cannot schedule new futures after shutdown.')
        if not self._loop.is_running():
            raise RuntimeError('Loop must be started before any function could be submitted.')
        if inspect.iscoroutinefunction(fn):
            coroutine = fn(*args, **kwargs)
            return asyncio.run_coroutine_threadsafe(coroutine, self._loop)
        else:
            func = functools.partial(fn, *args, **kwargs)
            return self._loop.run_in_executor(self._thread_pool, func)

    def shutdown(self, wait=True):
        self._shutdown = True
        if wait:
            self._thread_pool.shutdown()


async def _call_behavior_async(rpc_event, state, behavior, argument, request_deserializer):
    context = _Context(rpc_event, state, request_deserializer)
    try:
        return await behavior(argument, context), True
    except Exception as e:
        with state.condition:
            if e not in state.rpc_errors:
                logging.exception(e)
                _abort(state, rpc_event.operation_call, StatusCode.unknown, _common.encode(e))
        return None, False


async def _unary_response_in_pool_async(rpc_event, state, behavior, argument_thunk, request_deserializer,
                                        response_serializer):
    argument = argument_thunk()
    if argument is not None:
        response, proceed = await _call_behavior_async(rpc_event, state, behavior, argument, request_deserializer)
        if proceed:
            serialized_response = _serialize_response(rpc_event, state, response, response_serializer)
            if serialized_response is not None:
                _status(rpc_event, state, serialized_response)


def _handle_unary_unary(rpc_event, state, method_handler, default_thread_pool):
    unary_request = _unary_request(rpc_event, state, method_handler.request_deserializer)
    thread_pool = _select_thread_pool_for_behavior(method_handler.unary_unary, default_thread_pool)
    if asyncio.iscoroutinefunction(method_handler.unary_unary):
        return thread_pool.submit(_unary_response_in_pool_async, rpc_event, state, method_handler.unary_unary,
                                  unary_request, method_handler.request_deserializer,
                                  method_handler.response_serializer)
    else:
        return thread_pool.submit(_unary_response_in_pool, rpc_event, state, method_handler.unary_unary, unary_request,
                                  method_handler.request_deserializer, method_handler.response_serializer)


_server._handle_unary_unary = _handle_unary_unary



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import asyncio
import time
import traceback

from notification_service.base_notification import BaseEvent
from notification_service.event_storage import BaseEventStorage, DbEventStorage
from notification_service.high_availability import NotificationServerHaManager
from notification_service.mongo_event_storage import MongoEventStorage
from notification_service.proto import notification_service_pb2_grpc, notification_service_pb2
from notification_service.util.utils import event_to_proto, event_list_to_proto, member_to_proto, event_proto_to_event, \
    proto_to_member, count_list_to_proto
from notification_service.util.db import extract_db_engine_from_uri, DBType, parse_mongo_uri


class NotificationService(notification_service_pb2_grpc.NotificationServiceServicer):

    def __init__(self, storage: BaseEventStorage):
        self.storage = storage
        self.notification_conditions = {}
        self.lock = asyncio.Lock()
        self.write_condition = asyncio.Condition()

    @classmethod
    def from_storage_uri(cls, storage_uri: str) -> 'NotificationService':
        """
        Construct the notification service with the given storage uri
        :param storage_uri: uri of the backend storage to use
        :type storage_uri: str
        :rtype: NotificationService
        """
        db_engine = extract_db_engine_from_uri(storage_uri)
        if DBType.value_of(db_engine) == DBType.MONGODB:
            username, password, host, port, db = parse_mongo_uri(storage_uri)
            storage = MongoEventStorage(host=host,
                                        port=int(port),
                                        username=username,
                                        password=password,
                                        db=db)
            return cls(storage=storage)
        else:
            return cls(storage=DbEventStorage(storage_uri))

    def start(self):
        pass

    def stop(self):
        pass

    @asyncio.coroutine
    def sendEvent(self, request, context):
        try:
            return self._send_event(request)
        except Exception as e:
            print(e)
            traceback.print_stack()
            return notification_service_pb2.SendEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _send_event(self, request):
        event_proto = request.event
        event = BaseEvent(
            key=event_proto.key,
            value=event_proto.value,
            event_type=None if event_proto.event_type == "" else event_proto.event_type,
            context=None if event_proto.context == "" else event_proto.context,
            namespace=None if event_proto.namespace == "" else event_proto.namespace,
            sender=None if event_proto.sender == "" else event_proto.sender
        )
        uuid = request.uuid
        enable_idempotence = request.enable_idempotence
        key = event.key
        return_msg = ''
        # Lock conditions dict for get/check/update of key
        await self.lock.acquire()
        if self.notification_conditions.get(key) is None:
            self.notification_conditions.update({(key, asyncio.Condition())})
        # Release lock after check/update key of notification conditions dict
        self.lock.release()
        async with self.notification_conditions.get(key), self.write_condition:
            if enable_idempotence and self.storage.get_event_by_uuid(uuid) is not None:
                return_msg = 'Ignored because event already exists.'
            else:
                event: BaseEvent = self.storage.add_event(event, uuid)
            self.notification_conditions.get(key).notify_all()
            self.write_condition.notify_all()

        result_event_proto = event_to_proto(event)
        return notification_service_pb2.SendEventsResponse(
            event=result_event_proto,
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg=return_msg)

    @asyncio.coroutine
    def listEvents(self, request, context):
        try:
            return self._list_events(request)
        except Exception as e:
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _list_events(self, request):
        keys = request.keys
        event_type = request.event_type
        start_time = request.start_time
        start_version = request.start_version
        namespace = None if request.namespace == '' else request.namespace
        sender = None if request.sender == '' else request.sender
        timeout_seconds = request.timeout_seconds

        if timeout_seconds == 0:
            event_models = self._query_events(keys, event_type, start_time, start_version, namespace, sender)
            event_proto_list = event_list_to_proto(event_models)
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                events=event_proto_list)
        else:
            start = time.time()
            # Lock conditions dict for get/check/update of key
            await self.lock.acquire()
            for key in keys:
                if self.notification_conditions.get(key) is None:
                    self.notification_conditions.update({(key, asyncio.Condition())})
            # Release lock after check/update key of notification conditions dict
            self.lock.release()
            event_models = []
            if len(keys) == 1:
                key = keys[0]
                condition = self.notification_conditions.get(key)
            else:
                condition = self.write_condition
            async with condition:
                while time.time() - start < timeout_seconds and len(event_models) == 0:
                    try:
                        await asyncio.wait_for(condition.wait(),
                                               timeout_seconds - time.time() + start)
                        event_models = self._query_events(
                            keys, event_type, start_time, start_version, namespace, sender)
                    except asyncio.TimeoutError:
                        pass
                if len(event_models) == 0:
                    event_models = self._query_events(
                        keys, event_type, start_time, start_version, namespace, sender)
            event_proto_list = event_list_to_proto(event_models)
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                events=event_proto_list)

    def _query_events(self, keys, event_type, start_time, start_version, namespace, sender):
        return self.storage.list_events(keys, start_version, event_type, start_time, namespace, sender)

    def countEvents(self, request, context):
        keys = request.keys
        event_type = request.event_type
        start_time = request.start_time
        start_version = request.start_version
        namespace = None if request.namespace == '' else request.namespace
        sender = None if request.sender == '' else request.sender
        try:
            event_counts = self.storage.count_events(keys, start_version, event_type, start_time, namespace, sender)
            event_count, count_proto_list = count_list_to_proto(event_counts)
            return notification_service_pb2.CountEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                event_count=event_count,
                sender_event_counts=count_proto_list)
        except Exception as e:
            return notification_service_pb2.CountEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    @asyncio.coroutine
    def listAllEvents(self, request, context):
        try:
            return self._list_all_events(request)
        except Exception as e:
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _list_all_events(self, request):
        start_time = request.start_time
        start_version = request.start_version
        end_version = request.end_version
        timeout_seconds = request.timeout_seconds
        if 0 == timeout_seconds:
            event_models = self._query_all_events(start_time, start_version, end_version)
            event_proto_list = event_list_to_proto(event_models)
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                events=event_proto_list)
        else:
            start = time.time()
            event_models = self._query_all_events(start_time, start_version, end_version)
            async with self.write_condition:
                while time.time() - start < timeout_seconds and len(event_models) == 0:
                    try:
                        await asyncio.wait_for(self.write_condition.wait(),
                                               timeout_seconds - time.time() + start)
                        event_models = self._query_all_events(
                            start_time, start_version, end_version)
                    except asyncio.TimeoutError:
                        pass
            event_proto_list = event_list_to_proto(event_models)
            return notification_service_pb2.ListEventsResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                events=event_proto_list)

    def _query_all_events(self, start_time, start_version, end_version):
        if start_version > 0:
            return self.storage.list_all_events_from_version(start_version, end_version)
        else:
            return self.storage.list_all_events(start_time)

    @asyncio.coroutine
    def getLatestVersionByKey(self, request, context):
        try:
            return self._get_latest_version_by_key(request)
        except Exception as e:
            return notification_service_pb2.GetLatestVersionResponse(
                return_code=str(notification_service_pb2.ReturnStatus.ERROR), return_msg=str(e))

    async def _get_latest_version_by_key(self, request):
        key = request.key
        namespace = request.namespace
        timeout_seconds = request.timeout_seconds
        if 0 == timeout_seconds:
            latest_version = self._query_latest_version_by_key(key, namespace)
            return notification_service_pb2.GetLatestVersionResponse(
                return_code=str(notification_service_pb2.ReturnStatus.SUCCESS),
                return_msg='',
                version=latest_version)
        else:
            start = time.time()
            latest_version = self._query_latest_version_by_key(key, namespace)
            async with self.write_condition:
                while time.time() - start < timeout_seconds and latest_version == 0:
                    try:
                        await asyncio.wait_for(self.write_condition.wait(), timeout_seconds - time.time() + start)
                        latest_version = self._query_latest_version_by_key(key, namespace)
                    except asyncio.TimeoutError:
                        pass
            return notification_service_pb2.ListEventsResponse(
                return_code=str(notification_service_pb2.ReturnStatus.SUCCESS),
                return_msg='',
                version=latest_version)

    def _query_latest_version_by_key(self, key, namespace):
        if len(namespace) == 0:
            namespace = None
        return self.storage.get_latest_version(key=key, namespace=namespace)

    @asyncio.coroutine
    def notify(self, request, context):
        try:
            return self._notify(request)
        except Exception as e:
            return notification_service_pb2.CommonResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _notify(self, request):
        for notify in request.notifies:
            if notify.key in self.notification_conditions:
                async with self.notification_conditions.get(notify.key):
                    self.notification_conditions.get(notify.key).notify_all()
        async with self.write_condition:
            self.write_condition.notify_all()
        return notification_service_pb2.CommonResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='')

    @asyncio.coroutine
    def listMembers(self, request, context):
        try:
            return self._list_members(request)
        except Exception as e:
            return notification_service_pb2.ListMembersResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _list_members(self, request):
        # this method is used in HA mode, so we just return an empty list here.
        return notification_service_pb2.ListMembersResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='',
            members=[])

    @asyncio.coroutine
    def notifyNewMember(self, request, context):
        try:
            return self._notify_new_member(request)
        except Exception as e:
            return notification_service_pb2.CommonResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _notify_new_member(self, request):
        # this method is used in HA mode, so we just return here.
        return notification_service_pb2.CommonResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='')

    @asyncio.coroutine
    def registerClient(self, request, context):
        try:
            return self._register_client(request)
        except Exception as e:
            return notification_service_pb2.RegisterClientResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _register_client(self, request):
        client_id = self.storage.register_client(request.client_meta.namespace, request.client_meta.sender)
        return notification_service_pb2.RegisterClientResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='',
            client_id=client_id)

    @asyncio.coroutine
    def deleteClient(self, request, context):
        try:
            return self._delete_client(request)
        except Exception as e:
            return notification_service_pb2.CommonResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e))

    async def _delete_client(self, request):
        self.storage.delete_client(request.client_id)
        return notification_service_pb2.CommonResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='')

    @asyncio.coroutine
    def isClientExists(self, request, context):
        try:
            return self._is_client_exists(request)
        except Exception as e:
            return notification_service_pb2.isClientExistsResponse(
                return_code=notification_service_pb2.ReturnStatus.ERROR, return_msg=str(e), is_exists=False)

    async def _is_client_exists(self, request):
        is_exists = self.storage.is_client_exists(request.client_id)
        return notification_service_pb2.isClientExistsResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='',
            is_exists=is_exists)


class HighAvailableNotificationService(NotificationService):

    def __init__(self,
                 storage,
                 ha_manager,
                 server_uri,
                 ha_storage,
                 ttl_ms: int = 10000,
                 min_notify_interval_ms: int = 100):
        super(HighAvailableNotificationService, self).__init__(storage)
        self.server_uri = server_uri
        self.ha_storage = ha_storage
        self.ttl_ms = ttl_ms
        self.min_notify_interval_ms = min_notify_interval_ms
        self.ha_manager = ha_manager  # type: NotificationServerHaManager
        self.member_updated_condition = asyncio.Condition()

    def start(self):
        self.ha_manager.start(self.server_uri,
                              self.ha_storage,
                              self.ttl_ms,
                              self.min_notify_interval_ms,
                              self.member_updated_condition)

    def stop(self):
        self.ha_manager.stop()

    async def _send_event(self, request):
        response = await super(HighAvailableNotificationService, self)._send_event(request)
        try:
            if response.return_code == notification_service_pb2.ReturnStatus.SUCCESS:
                self.ha_manager.notify_others(event_proto_to_event(response.event))
        finally:
            return response

    async def _list_members(self, request):
        timeout_seconds = request.timeout_seconds
        if 0 == timeout_seconds:
            return notification_service_pb2.ListMembersResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                members=[member_to_proto(member) for member in self.ha_manager.get_living_members()])
        else:
            start = time.time()
            members = self.ha_manager.get_living_members()
            async with self.member_updated_condition:
                while time.time() - start < timeout_seconds:
                    try:
                        await asyncio.wait_for(self.member_updated_condition.wait(),
                                               timeout_seconds - time.time() + start)
                        members = self.ha_manager.get_living_members()
                        break
                    except asyncio.TimeoutError:
                        pass
            return notification_service_pb2.ListMembersResponse(
                return_code=notification_service_pb2.ReturnStatus.SUCCESS,
                return_msg='',
                members=[member_to_proto(member) for member in members])

    async def _notify_new_member(self, request):
        self.ha_manager.add_living_member(proto_to_member(request.member))
        async with self.member_updated_condition:
            self.member_updated_condition.notify_all()
        return notification_service_pb2.CommonResponse(
            return_code=notification_service_pb2.ReturnStatus.SUCCESS,
            return_msg='')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import yaml


class NotificationServerConfig(object):
    def __init__(self, config_file):
        self.config_file = config_file
        self._port = 50052
        self._db_uri = None
        self._enable_ha = False
        self._ha_ttl_ms = None
        self._advertised_uri = None
        self._parse_config()

    def _parse_config(self):
        with open(self.config_file, 'r') as f:
            yaml_config = yaml.load(f, Loader=yaml.FullLoader)
            if 'server_port' in yaml_config:
                self._port = yaml_config['server_port']
            if 'db_uri' in yaml_config:
                self._db_uri = yaml_config['db_uri']
            if 'enable_ha' in yaml_config:
                self._enable_ha = yaml_config['enable_ha']
            if 'ha_ttl_ms' in yaml_config:
                self._ha_ttl_ms = yaml_config['ha_ttl_ms']
            if 'advertised_uri' in yaml_config:
                self._advertised_uri = yaml_config['advertised_uri']

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @property
    def db_uri(self):
        return self._db_uri

    @db_uri.setter
    def db_uri(self, db_uri):
        self._db_uri = db_uri

    @property
    def enable_ha(self):
        return self._enable_ha

    @enable_ha.setter
    def enable_ha(self, enable_ha):
        self._enable_ha = enable_ha

    @property
    def ha_ttl_ms(self):
        return self._ha_ttl_ms

    @ha_ttl_ms.setter
    def ha_ttl_ms(self, ha_ttl_ms):
        self._ha_ttl_ms = ha_ttl_ms

    @property
    def advertised_uri(self):
        return self._advertised_uri

    @advertised_uri.setter
    def advertised_uri(self, advertised_uri):
        self._advertised_uri = advertised_uri



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import socket
from notification_service.base_notification import BaseEvent, Member, SenderEventCount
from notification_service.proto import notification_service_pb2

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)


def event_to_proto(event: BaseEvent):
    result_event_proto = notification_service_pb2.EventProto(key=event.key,
                                                             version=event.version,
                                                             value=event.value,
                                                             event_type=event.event_type,
                                                             create_time=event.create_time,
                                                             namespace=event.namespace,
                                                             context=event.context,
                                                             sender=event.sender)
    return result_event_proto


def count_to_proto(count: SenderEventCount):
    sender_event_count_proto = notification_service_pb2.SenderEventCountProto(sender=count.sender,
                                                                              event_count=count.event_count)
    return sender_event_count_proto


def event_list_to_proto(event_list):
    event_proto_list = []
    for event_model in event_list:
        event_proto = event_to_proto(event_model)
        event_proto_list.append(event_proto)
    return event_proto_list


def count_list_to_proto(count_list):
    event_count = 0
    count_proto_list = []
    for count in count_list:
        event_count += count.event_count
        count_proto_list.append(count_to_proto(count))
    return event_count, count_proto_list


def event_proto_to_event(event_proto):
    return BaseEvent(key=event_proto.key,
                     value=event_proto.value,
                     event_type=event_proto.event_type,
                     version=event_proto.version,
                     create_time=event_proto.create_time,
                     context=event_proto.context,
                     namespace=event_proto.namespace,
                     sender=event_proto.sender)


def event_count_proto_to_event_count(event_count_proto):
    return SenderEventCount(sender=event_count_proto.sender,
                            event_count=event_count_proto.event_count)


def event_model_to_event(event_model):
    return BaseEvent(
        key=event_model.key,
        value=event_model.value,
        event_type=event_model.event_type,
        version=event_model.version,
        create_time=event_model.create_time,
        context=event_model.context,
        namespace=event_model.namespace,
        sender=event_model.sender
    )


def count_result_to_sender_event_count(count_result):
    return SenderEventCount(
        sender=count_result.sender,
        event_count=count_result.event_count
    )


def member_to_proto(member: Member):
    return notification_service_pb2.MemberProto(
        version=member.version, server_uri=member.server_uri, update_time=member.update_time)


def proto_to_member(member_proto):
    return Member(member_proto.version, member_proto.server_uri, member_proto.update_time)


def sleep_and_detecting_running(interval_ms, is_running_callable, min_interval_ms=500):
    start_time = time.monotonic() * 1000
    while is_running_callable() and time.monotonic() * 1000 < start_time + interval_ms:
        remaining = time.monotonic() * 1000 - start_time
        if remaining > min_interval_ms:
            time.sleep(min_interval_ms / 1000)
        else:
            time.sleep(remaining / 1000)


def get_ip_addr() -> str:
    """
    Get ip address of localhost by UDP socket.
    :return: ip address of localhost
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    except OSError:
        ip = '127.0.0.1'
    finally:
        s.close()
        return ip



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import time
import socket
from collections import Iterable
from typing import Union, Tuple

from mongoengine import connect
from notification_service.event_storage import BaseEventStorage
from notification_service.base_notification import BaseEvent
from notification_service.mongo_notification import MongoEvent, MongoClientModel


class MongoEventStorage(BaseEventStorage):
    def __init__(self, *args, **kwargs):
        self.db_conn = self.setup_connection(**kwargs)
        self.db = None
        self.server_ip = socket.gethostbyname(socket.gethostname())

    def setup_connection(self, **kwargs):
        db_conf = {
            "host": kwargs.get("host"),
            "port": kwargs.get("port"),
            "db": kwargs.get("db"),
        }
        self.db = db_conf.get('db')
        username = kwargs.get("username", None)
        password = kwargs.get("password", None)
        authentication_source = kwargs.get("authentication_source", "admin")
        if (username or password) and not (username and password):
            raise Exception("Please provide valid username and password")
        if username and password:
            db_conf.update({
                "username": username,
                "password": password,
                "authentication_source": authentication_source
            })
        return connect(**db_conf)

    def get_latest_version(self, key: str, namespace: str = None):
        mongo_events = MongoEvent.get_by_key(key, 0, 1, "-version")
        if not mongo_events:
            return 0
        return mongo_events[0].version

    def add_event(self, event: BaseEvent, uuid: str):
        kwargs = {
            "server_ip": self.server_ip,
            "create_time": int(time.time() * 1000),
            "event_type": event.event_type,
            "key": event.key,
            "value": event.value,
            "context": event.context,
            "namespace": event.namespace,
            "sender": event.sender,
            "uuid": uuid
        }
        mongo_event = MongoEvent(**kwargs)
        mongo_event.save()
        mongo_event.reload()
        event.create_time = mongo_event.create_time
        event.version = mongo_event.version
        return event

    def list_events(self,
                    key: Union[str, Tuple[str]],
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    namespace: str = None,
                    sender: str = None):
        key = None if key == "" else key
        version = None if version == 0 else version
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        res = MongoEvent.get_base_events(key, version, event_type, start_time, namespace, sender)
        return res

    def count_events(self,
                     key: Union[str, Tuple[str]],
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     namespace: str = None,
                     sender: str = None):
        key = None if key == "" else key
        version = None if version == 0 else version
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        res = MongoEvent.count_base_events(key, version, event_type, start_time, namespace, sender)
        return res

    def list_all_events(self, start_time: int):
        res = MongoEvent.get_base_events_by_time(start_time)
        return res

    def list_all_events_from_version(self, start_version: int, end_version: int = None):
        res = MongoEvent.get_base_events_by_version(start_version, end_version)
        return res

    def register_client(self, namespace: str = None, sender: str = None) -> int:
        client = MongoClientModel(namespace=namespace, sender=sender, create_time=int(time.time() * 1000))
        client.save()
        return client.id

    def delete_client(self, client_id):
        client_to_delete = MongoClientModel.objects(id=client_id).first()
        if client_to_delete is None:
            raise Exception("You are trying to delete an non-existing notification client!")
        client_to_delete.is_deleted = True
        client_to_delete.save()

    def is_client_exists(self, client_id) -> bool:
        client_to_check = MongoClientModel.objects(id=client_id, is_deleted__ne=True).first()
        return client_to_check is not None

    def get_event_by_uuid(self, uuid: str):
        return MongoEvent.get_event_by_uuid(uuid)

    def clean_up(self):
        MongoEvent.delete_by_client(self.server_ip)
        if self.db is not None:
            self.db_conn.drop_database(self.db)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import contextlib
import re
import time
import urllib.parse
from collections import Iterable
from enum import Enum
from functools import wraps
from typing import Tuple, Union

from sqlalchemy import create_engine, Column, String, BigInteger, Text, Integer, Boolean, func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from notification_service.base_notification import BaseEvent, Member, ANY_CONDITION
from notification_service.util.utils import event_model_to_event, count_result_to_sender_event_count

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

# use sqlite by default for testing
SQL_ALCHEMY_DB_FILE = 'notification_service.db'
SQL_ALCHEMY_CONN = "sqlite:///" + SQL_ALCHEMY_DB_FILE
engine = None
Session = None

"""
List of SQLAlchemy database engines for notification backend storage.
"""

MYSQL = 'mysql'
SQLITE = 'sqlite'
POSTGRES = 'postgresql'
MSSQL = 'mssql'
MONGODB = 'mongodb'

DATABASE_ENGINES = [
    MYSQL,
    SQLITE,
    POSTGRES,
    MSSQL,
    MONGODB,
]


class DBType(str, Enum):
    SQLITE = "sql_lite"
    MYSQL = "mysql"
    MONGODB = "mongodb"

    @staticmethod
    def value_of(db_type):
        if db_type in ('SQL_LITE', 'sql_lite', 'sqlite'):
            return DBType.SQLITE
        elif db_type in ('MYSQL', 'mysql'):
            return DBType.MYSQL
        elif db_type in ('MONGODB', 'mongodb'):
            return DBType.MONGODB
        else:
            raise NotImplementedError


def extract_db_engine_from_uri(db_uri):
    """
    Parse specified database URI to extract database type. Confirm extracted database engine is
    supported. If database driver is specified, confirm driver passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count('+')

    """validates scheme parsed from DB URI is supported"""
    if scheme_plus_count == 0:
        db_engine = scheme
    elif scheme_plus_count == 1:
        db_engine, _ = scheme.split('+')
    else:
        error_msg = "Invalid database URI: '%s'." % db_uri
        raise Exception(error_msg)

    """validates db_engine parsed from DB URI is supported"""
    if db_engine not in DATABASE_ENGINES:
        error_msg = "Invalid database engine: '%s'." % db_engine
        raise Exception(error_msg)

    return db_engine


def parse_mongo_uri(db_uri):
    """
    Parse MongoDB URI-style string to split up and return credentials

    Args:
        db_uri (string): MongoDB URI-style string
    Return:

    """
    regex_str = r'^(?P<schema>(mongodb:(?:\/{2})?))((?P<user>\w+?):(?P<pwd>(\w+?))@|:@?)(?P<host>(\S+?)):(?P<port>(\d+))(\/(?P<db>(\S+?)))$'
    pattern = re.compile(regex_str)
    m = pattern.match(db_uri)
    if m is None:
        raise Exception('The URI of MongoDB is invalid')
    return m.group('user'), m.group('pwd'), m.group('host'), m.group('port'), m.group('db')


def prepare_db(user_engine=None, user_session=None, print_sql=False):
    global engine
    global Session
    if user_engine is not None and user_session is not None:
        engine = user_engine
        Session = user_session
    if engine is None or Session is None:
        engine_args = {'encoding': "utf-8"}
        if print_sql:
            engine_args['echo'] = True
        engine = create_engine(SQL_ALCHEMY_CONN, **engine_args)
        Session = scoped_session(
            sessionmaker(autocommit=False,
                         autoflush=False,
                         bind=engine,
                         expire_on_commit=False))


@contextlib.contextmanager
def create_session():
    prepare_db()
    session = Session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def provide_session(func):
    """
    Function decorator that provides a session if it isn't provided.
    If you want to reuse a session or run the function as part of a
    database transaction, you pass it to the function, if not this wrapper
    will create one and close it for you.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        arg_session = 'session'

        func_params = func.__code__.co_varnames
        session_in_args = arg_session in func_params and \
                          func_params.index(arg_session) < len(args)
        session_in_kwargs = arg_session in kwargs

        if session_in_kwargs or session_in_args:
            return func(*args, **kwargs)
        else:
            with create_session() as session:
                kwargs[arg_session] = session
                return func(*args, **kwargs)

    return wrapper


Base = declarative_base()


class EventModel(Base):
    __tablename__ = "event_model"
    version = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    key = Column(String(1024), nullable=False)
    value = Column(Text())
    event_type = Column(String(1024), server_default="UNDEFINED")
    context = Column(Text())
    namespace = Column(String(1024))
    sender = Column(String(1024))
    create_time = Column(BigInteger(), nullable=False)
    uuid = Column(String(40), nullable=False, unique=True)

    @staticmethod
    @provide_session
    def add_event(event: BaseEvent, uuid, session=None):
        event_model = EventModel()
        event_model.key = event.key
        event_model.value = event.value
        event_model.event_type = event.event_type
        event_model.context = event.context
        event_model.namespace = event.namespace
        event_model.create_time = int(time.time() * 1000)
        event_model.uuid = uuid
        event_model.sender = event.sender
        session.add(event_model)
        session.commit()
        return event_model_to_event(event_model)

    @staticmethod
    @provide_session
    def list_events(key: Union[str, Tuple[str]],
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    namespace: str = None,
                    sender: str = None,
                    session=None):
        key = None if key == "" else key
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        if key is None:
            raise Exception('key cannot be empty.')

        conditions = []
        if event_type is not None and event_type != ANY_CONDITION:
            conditions.append(EventModel.event_type == event_type)
        if start_time is not None and start_time > 0:
            conditions.append(EventModel.create_time >= start_time)
        if namespace is not None and ANY_CONDITION != namespace:
            conditions.append(EventModel.namespace == namespace)
        if sender is not None and ANY_CONDITION != sender:
            conditions.append(EventModel.sender == sender)
        if version > 0:
            conditions.append(EventModel.version > version)
        if ANY_CONDITION not in key:
            conditions.append(EventModel.key.in_(key))
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def count_events(key: Union[str, Tuple[str]],
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     namespace: str = None,
                     sender: str = None,
                     session=None):
        key = None if key == "" else key
        event_type = None if event_type == "" else event_type
        namespace = None if namespace == "" else namespace
        sender = None if sender == "" else sender
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        if key is None:
            raise Exception('key cannot be empty.')
        conditions = []
        if event_type is not None and event_type != ANY_CONDITION:
            conditions.append(EventModel.event_type == event_type)
        if start_time is not None and start_time > 0:
            conditions.append(EventModel.create_time >= start_time)
        if namespace is not None and ANY_CONDITION != namespace:
            conditions.append(EventModel.namespace == namespace)
        if sender is not None and ANY_CONDITION != sender:
            conditions.append(EventModel.sender == sender)
        if version > 0:
            conditions.append(EventModel.version > version)
        if ANY_CONDITION not in key:
            conditions.append(EventModel.key.in_(key))

        count_results = session.query(EventModel.sender, func.count('*').label('event_count')).filter(*conditions).group_by(
            EventModel.sender)
        return [count_result_to_sender_event_count(count_result) for count_result in count_results]

    @staticmethod
    @provide_session
    def list_all_events(start_time: int, session=None):
        conditions = [
            EventModel.create_time >= start_time
        ]
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def list_all_events_from_version(start_version: int, end_version: int = None, session=None):
        conditions = [
            EventModel.version > start_version
        ]
        if end_version is not None and end_version > 0:
            conditions.append(EventModel.version <= end_version)
        event_model_list = session.query(EventModel).filter(*conditions).all()
        return [event_model_to_event(event_model) for event_model in event_model_list]

    @staticmethod
    @provide_session
    def sync_event(event: BaseEvent, uuid, session=None):
        event_model = EventModel()
        event_model.key = event.key
        event_model.value = event.value
        event_model.event_type = event.event_type
        event_model.context = event.context
        event_model.namespace = event.namespace
        event_model.create_time = event.create_time
        event_model.uuid = uuid
        event_model.sender = event.sender
        session.add(event_model)
        session.commit()
        return event_model_to_event(event_model)

    @staticmethod
    @provide_session
    def get_latest_version(session=None):
        event = session.query(EventModel).order_by(EventModel.version.desc()) \
            .limit(1).first()
        if event is not None:
            return event.version
        else:
            return 0

    @staticmethod
    @provide_session
    def get_event_by_uuid(uuid, session=None):
        conditions = [
            EventModel.uuid == uuid
        ]
        return session.query(EventModel).filter(*conditions).first()

    @staticmethod
    def create_table(db_conn=None):
        if db_conn is not None:
            global SQL_ALCHEMY_CONN
            SQL_ALCHEMY_CONN = db_conn
        prepare_db()
        if not engine.dialect.has_table(engine, EventModel.__tablename__):
            Base.metadata.create_all(engine)

    @staticmethod
    @provide_session
    def cleanup(session=None):
        session.query(EventModel).delete()
        session.commit()


class ClientModel(Base):
    __tablename__ = "notification_client"
    id = Column(BigInteger().with_variant(Integer, "sqlite"), autoincrement=True, primary_key=True)
    namespace = Column(String(1024))
    sender = Column(String(1024))
    create_time = Column(BigInteger)
    is_deleted = Column(Boolean, default=False)

    @staticmethod
    @provide_session
    def register_client(namespace: str = None, sender: str = None, session=None):
        client = ClientModel()
        client.namespace = namespace
        client.sender = sender
        client.create_time = int(time.time() * 1000)
        session.add(client)
        session.commit()
        return client.id

    @staticmethod
    @provide_session
    def delete_client(client_id, session=None):
        client_to_delete = session.query(ClientModel).filter(ClientModel.id == client_id).first()
        if client_to_delete is None:
            raise Exception("You are trying to delete an non-existing notification client!")
        client_to_delete.is_deleted = True
        session.flush()

    @staticmethod
    @provide_session
    def is_client_exists(client_id, session=None):
        client_to_check = session.query(ClientModel).filter(ClientModel.id == client_id,
                                                            ClientModel.is_deleted.is_(False)).first()
        return client_to_check is not None

    @staticmethod
    def create_table(db_conn=None):
        if db_conn is not None:
            global SQL_ALCHEMY_CONN
            SQL_ALCHEMY_CONN = db_conn
        prepare_db()
        if not engine.dialect.has_table(engine, ClientModel.__tablename__):
            Base.metadata.create_all(engine)

    @staticmethod
    @provide_session
    def cleanup(session=None):
        session.query(ClientModel).delete()
        session.commit()


class MemberModel(Base):
    __tablename__ = "member_model"
    id = Column(BigInteger().with_variant(Integer, "sqlite"), primary_key=True)
    version = Column(BigInteger(), nullable=False)
    server_uri = Column(String(767), nullable=False, unique=True)
    update_time = Column(BigInteger(), nullable=False)
    uuid = Column(String(128), nullable=False, unique=True)

    @staticmethod
    def create_table(db_conn=None):
        if db_conn is not None:
            global SQL_ALCHEMY_CONN
            SQL_ALCHEMY_CONN = db_conn
        prepare_db()
        if not engine.dialect.has_table(engine, MemberModel.__tablename__):
            Base.metadata.create_all(engine)

    @staticmethod
    @provide_session
    def cleanup(session=None):
        session.query(MemberModel).delete()
        session.commit()

    @staticmethod
    @provide_session
    def update_member(server_uri, server_uuid, session=None):
        member = session.query(MemberModel) \
            .filter(MemberModel.server_uri == server_uri).first()
        if member is None:
            member = MemberModel()
            member.version = 1
            member.server_uri = server_uri
            member.update_time = time.time_ns() / 1000000
            member.uuid = server_uuid
            session.add(member)
        else:
            if member.uuid != server_uuid:
                raise Exception("The server uri '%s' is already exists in the storage!" %
                                server_uri)
            member.version += 1
            member.update_time = time.time_ns() / 1000000
        session.commit()

    @staticmethod
    @provide_session
    def get_living_members(ttl, session=None):
        member_models = session.query(MemberModel) \
            .filter(MemberModel.update_time >= time.time_ns() / 1000000 - ttl) \
            .all()
        return [Member(m.version, m.server_uri, int(m.update_time)) for m in member_models]

    @staticmethod
    @provide_session
    def get_dead_members(ttl, session=None):
        member_models = session.query(MemberModel) \
            .filter(MemberModel.update_time < time.time_ns() / 1000000 - ttl) \
            .all()
        return [Member(m.version, m.server_uri, int(m.update_time)) for m in member_models]

    @staticmethod
    @provide_session
    def delete_member(server_uri=None, server_uuid=None, session=None):
        conditions = []
        if server_uri:
            conditions.append(MemberModel.server_uri == server_uri)
        if server_uuid:
            conditions.append(MemberModel.uuid == server_uuid)
        if len(conditions) != 1:
            raise Exception("Please provide exactly one param, server_uri or server_uuid")
        member = session.query(MemberModel).filter(*conditions).first()
        if member is not None:
            session.delete(member)
        session.commit()

    @staticmethod
    @provide_session
    def clear_dead_members(ttl, session=None):
        session.query(MemberModel) \
            .filter(MemberModel.update_time < time.time_ns() / 1000000 - ttl) \
            .delete()
        session.commit()


def create_all_tables(db_conn=None):
    if db_conn is not None:
        global SQL_ALCHEMY_CONN
        SQL_ALCHEMY_CONN = db_conn
    prepare_db()
    Base.metadata.create_all(engine)


def drop_all_tables(db_conn=None):
    if db_conn is not None:
        global SQL_ALCHEMY_CONN
        SQL_ALCHEMY_CONN = db_conn
    prepare_db()
    Base.metadata.drop_all(engine)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Tuple

from mongoengine import (Document, IntField, StringField, SequenceField, LongField, BooleanField)
from notification_service.base_notification import BaseEvent, ANY_CONDITION, SenderEventCount


class MongoEvent(Document):
    server_ip = StringField()   # track server that the event belongs to
    key = StringField()
    value = StringField()
    event_type = StringField()
    version = SequenceField()  # use 'version' as the auto increase id
    create_time = LongField()
    context = StringField()
    namespace = StringField()
    sender = StringField()
    uuid = StringField()

    def to_dict(self):
        return {
            "key": self.key,
            "value": self.value,
            "event_type": self.event_type,
            "version": int(self.version),
            "create_time": self.create_time,
            "context": self.context,
            "namespace": self.namespace,
            "sender": self.sender
        }

    def to_base_event(self):
        base_event = BaseEvent(**self.to_dict())
        return base_event

    @classmethod
    def convert_to_base_events(cls, mongo_events: list = None):
        if not mongo_events:
            return []
        base_events = []
        for mongo_event in mongo_events:
            base_events.append(mongo_event.to_base_event())
        return base_events

    @classmethod
    def convert_to_event_counts(cls, mongo_counts: list = None):
        if not mongo_counts:
            return []
        event_counts = []
        for mongo_count in mongo_counts:
            event_counts.append(SenderEventCount(sender=mongo_count.sender, event_count=mongo_count.count))
        return event_counts

    @classmethod
    def get_base_events_by_version(cls, start_version: int, end_version: int = None):
        conditions = dict()
        conditions["version__gt"] = start_version
        if end_version is not None and end_version > 0:
            conditions["version__lte"] = end_version
        mongo_events = cls.objects(**conditions).order_by("version")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def get_base_events(cls,
                        key: Tuple[str],
                        version: int = None,
                        event_type: str = None,
                        start_time: int = None,
                        namespace: str = None,
                        sender: str = None):
        conditions = dict()
        if len(key) == 1:
            if ANY_CONDITION != key[0]:
                conditions["key"] = key[0]
        elif len(key) > 1:
            conditions["key__in"] = list(key)
        if version is not None and version > 0:
            conditions["version__gt"] = version
        if event_type is not None:
            if event_type != ANY_CONDITION:
                conditions["event_type"] = event_type
        if start_time is not None and start_time > 0:
            conditions["start_time_gte"] = start_time
        if ANY_CONDITION != namespace:
            conditions["namespace"] = namespace
        if sender is not None and ANY_CONDITION != sender:
            conditions["sender"] = sender
        mongo_events = cls.objects(**conditions).order_by("version")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def count_base_events(cls,
                          key: Tuple[str],
                          version: int = None,
                          event_type: str = None,
                          start_time: int = None,
                          namespace: str = None,
                          sender: str = None):
        conditions = dict()
        if len(key) == 1:
            if ANY_CONDITION != key[0]:
                conditions["key"] = key[0]
        elif len(key) > 1:
            conditions["key__in"] = list(key)
        if version is not None and version > 0:
            conditions["version__gt"] = version
        if event_type is not None:
            if event_type != ANY_CONDITION:
                conditions["event_type"] = event_type
        if start_time is not None and start_time > 0:
            conditions["start_time_gte"] = start_time
        if ANY_CONDITION != namespace:
            conditions["namespace"] = namespace
        if sender is not None and ANY_CONDITION != sender:
            conditions["sender"] = sender
        mongo_counts = cls.objects(**conditions).aggregate([{"$group": {"_id": "$sender", "count": {"$sum": 1}}}])
        return cls.convert_to_event_counts(mongo_counts)

    @classmethod
    def get_base_events_by_time(cls, create_time: int = None):
        mongo_events = cls.objects(create_time__gte=create_time).order_by("version")
        return cls.convert_to_base_events(mongo_events)

    @classmethod
    def get_by_key(cls, key: str = None, start: int = None, end: int = None, sort_key: str = "version"):
        if key:
            return cls.objects(key=key).order_by(sort_key)[start:end]
        else:
            raise Exception("key is empty, please provide valid key")

    @classmethod
    def delete_by_client(cls, server_ip):
        cls.objects(server_ip=server_ip).delete()

    @classmethod
    def get_event_by_uuid(cls, uuid):
        event = cls.objects(uuid=uuid).first()
        return event.to_base_event()


class MongoClientModel(Document):
    id = SequenceField(primary_key=True)
    namespace = StringField()
    sender = StringField()
    create_time = LongField()
    is_deleted = BooleanField(default=False)

    def __repr__(self):
        return '<Document Client ({}, {}, {}, {})>'.format(
            self.id,
            self.namespace,
            self.sender,
            self.is_deleted)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import abc
import time

from typing import List, Union, Tuple

UNDEFINED_EVENT_TYPE = "UNDEFINED"
ANY_CONDITION = "*"
DEFAULT_NAMESPACE = "default"


class BaseEvent(object):
    def __init__(self,
                 key: str,
                 value: str,
                 event_type: str = UNDEFINED_EVENT_TYPE,
                 version: int = None,
                 create_time: int = None,
                 context: str = None,
                 namespace: str = None,
                 sender: str = None):
        self.key = key
        self.value = value
        self.event_type = event_type
        self.version = version
        self.create_time = create_time
        self.context = context
        self.namespace = namespace
        self.sender = sender

    def __str__(self) -> str:
        return 'key:{0}, value:{1}, type:{2}, version:{3}, create_time:{4}, ' \
               'context: {5}, namespace: {6}, sender: {7}' \
            .format(self.key, self.value, self.event_type, self.version, self.create_time,
                    self.context, self.namespace, self.sender)

    def __eq__(self, other):
        if not isinstance(other, BaseEvent):
            return False
        return self.key == other.key and \
            self.value == other.value and \
            self.event_type == other.event_type and \
            self.version == other.version and \
            self.create_time == other.create_time and \
            self.context == other.context and \
            self.namespace == other.namespace and \
            self.sender == other.sender


class SenderEventCount(object):
    def __init__(self, sender: str, event_count: int):
        self.sender = sender
        self.event_count = event_count

    def __str__(self) -> str:
        return 'sender:{0}, event_count:{1}'.format(self.sender, self.event_count)

    def __eq__(self, other):
        if not isinstance(other, SenderEventCount):
            return False
        return self.sender == other.sender and \
               self.event_count == other.event_count


class EventWatcher(metaclass=abc.ABCMeta):
    """
    SignalWatcher is used to represent a standard event handler, which defines the
    logic related to signal notifications.
    """

    @abc.abstractmethod
    def process(self, events: List[BaseEvent]):
        pass


class EventWatcherHandle(metaclass=abc.ABCMeta):
    """
    The EventWatcherHandle used to stop the relevant watch thread.
    """

    @abc.abstractmethod
    def stop(self):
        pass


class BaseNotification(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def send_event(self, event: BaseEvent):
        """
        Send event to Notification Service.

        :param event: the event updated.
        :return: The created event which has version and create time.
        """
        pass

    @abc.abstractmethod
    def list_events(self,
                    key: Union[str, List[str]],
                    namespace: str = None,
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    sender: str = None) -> List[BaseEvent]:
        """
        List specific events in Notification Service.

        :param key: Key of the event for listening.
        :param namespace: Namespace of the event for listening.
        :param version: (Optional) The version of the events must greater than this version.
        :param event_type: (Optional) Type of the events.
        :param start_time: (Optional) Start time of the events.
        :param sender: The event sender.
        :return: The event list.
        """
        pass

    @abc.abstractmethod
    def start_listen_event(self,
                           key: Union[str, Tuple[str]],
                           watcher: EventWatcher,
                           namespace: str = None,
                           version: int = None,
                           event_type: str = None,
                           start_time: int = None,
                           sender: str = None) -> EventWatcherHandle:
        """
        Start listen specific `key` or `version` notifications in Notification Service.

        :param key: Key of notification for listening.
        :param namespace: Namespace of the event for listening.
        :param watcher: Watcher instance for listening.
        :param version: (Optional) The version of the events must greater than this version.
        :param event_type: (Optional) Type of the events for listening.
        :param start_time: (Optional) Start time of the events for listening.
        :param sender: The event sender.
        :return: The handle used to stop the listening.
        """
        pass

    @abc.abstractmethod
    def stop_listen_event(self, key: Union[str, Tuple[str]] = None,
                          namespace: str = None,
                          event_type: str = None,
                          sender: str = None
                          ):
        """
        Stop listen specific `key` notifications in Notification Service.

        :param key: Keys of notification for listening.
        :param namespace: Namespace of notification for listening.
        :param event_type: (Optional) Type of the events for listening.
        :param sender: The event sender.
        """
        pass

    @abc.abstractmethod
    def list_all_events(self,
                        start_time: int = None,
                        start_version: int = None,
                        end_version: int = None) -> List[BaseEvent]:
        """
        List specific `key` or `version` of events in Notification Service.

        :param start_time: (Optional) Start time of the events.
        :param start_version: (Optional) the version of the events must greater than the
                              start_version.
        :param end_version: (Optional) the version of the events must equal or less than the
                            end_version.
        :return: The event list.
        """
        pass

    @abc.abstractmethod
    def start_listen_events(self,
                            watcher: EventWatcher,
                            start_time=int(time.time() * 1000),
                            version: int = None) -> EventWatcherHandle:
        """
        Start listen all events.

        :param watcher: process event.
        :param start_time: (Optional) the earliest event time.
        :param version: (Optional) the start version of the event.
        :return: The handle used to stop the listening.
        """
        pass

    @abc.abstractmethod
    def stop_listen_events(self):
        """
        Stop the global listening threads.
        """
        pass

    @abc.abstractmethod
    def get_latest_version(self, key: str = None, namespace: str = None):
        """
        get latest event's version by key.
        :param key: Key of notification for listening.
        :param namespace: Namespace of notification for listening.
        :return: Version number of the specific key.
        """
        pass

    @abc.abstractmethod
    def close(self):
        """
        Clear resources and close the notification client.
        """


class Member(object):

    def __init__(self, version, server_uri, update_time):
        self.version = version
        self.server_uri = server_uri
        self.update_time = update_time



# coding:utf-8
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os

from setuptools import setup, find_packages

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_script():
    bin_dir = os.path.join(CURRENT_DIR, "bin")
    return [os.path.join("bin", filename) for filename in os.listdir(bin_dir)]


setup(
    name='notification_service',
    version='0.2.2',
    description='A Python package which provides stable notification service.',
    author='',
    author_email='flink.aiflow@gmail.com',
    url='https://github.com/alibaba/flink-ai-extended',
    packages=find_packages(exclude=['tests*']),
    scripts=get_script(),
    include_package_data=True,
    install_requires=["protobuf==3.15.6", "grpcio==1.34.0", "sqlalchemy>=1.3.18, <2", "pyyaml>=5.1, <5.4"]
)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import logging
import os
import threading
import time
import uuid
from collections import Iterable
from functools import wraps
from random import shuffle
from typing import Union, List, Tuple, Dict, Any

import grpc

from notification_service.base_notification import BaseNotification, EventWatcher, BaseEvent, EventWatcherHandle, \
    ANY_CONDITION, SenderEventCount
from notification_service.proto import notification_service_pb2_grpc
from notification_service.proto.notification_service_pb2 \
    import SendEventRequest, ListEventsRequest, EventProto, ReturnStatus, ListAllEventsRequest, \
    GetLatestVersionByKeyRequest, ListMembersRequest, RegisterClientRequest, ClientMeta, ClientIdRequest, \
    CountEventsRequest
from notification_service.util.utils import event_proto_to_event, proto_to_member, sleep_and_detecting_running, \
    event_count_proto_to_event_count

if not hasattr(time, 'time_ns'):
    time.time_ns = lambda: int(time.time() * 1e9)

NOTIFICATION_TIMEOUT_SECONDS = os.environ.get("NOTIFICATION_TIMEOUT_SECONDS", 5)
ALL_EVENTS_KEY = "_*"
ENABLE_IDEMPOTENCE_CONFIG = 'enable.idempotence'
CLIENT_ID = 'client.id'
INITIAL_SEQUENCE_NUMBER = 'initial.sequence.number'
GRPC_MAX_RECEIVE_MESSAGE_LENGTH = 'grpc.max_receive_message_length'


class ThreadEventWatcherHandle(EventWatcherHandle):

    def __init__(self,
                 thread: threading.Thread,
                 thread_key: Any,
                 notification_client: 'NotificationClient'):
        self._thread = thread
        self._thread_key = thread_key
        self._notification_client = notification_client

    def stop(self):
        self._thread._flag = False
        self._thread.join()
        self._notification_client.lock.acquire()
        try:
            self._notification_client.threads[self._thread_key].remove(self._thread)
        finally:
            self._notification_client.lock.release()


class SequenceNumberManager(object):

    def __init__(self, seq_num: int = 0):
        self._lock = threading.RLock()
        self._seq_num = seq_num

    def increment_sequence_number(self):
        with self._lock:
            self._seq_num += 1

    def get_sequence_number(self):
        return self._seq_num


class NotificationClient(BaseNotification):
    """
    NotificationClient is the notification client.
    """

    def __init__(self,
                 server_uri: str,
                 default_namespace: str = None,
                 enable_ha: bool = False,
                 list_member_interval_ms: int = 5000,
                 retry_interval_ms: int = 1000,
                 retry_timeout_ms: int = 10000,
                 sender: str = None,
                 properties: Dict[str, str] = None):
        """
        The constructor of the NotificationClient.
        The maximum message length that the NotificationClient can receive is set to unlimited for now.

        :param server_uri: Target server uri/uris. If multiple uris separated by "," were provided,
                           the HA mode would be turned on, regardless of the parameter `enable_ha`.
        :param default_namespace: The default namespace that this client is working on.
        :param enable_ha: Enable high-available functionality.
        :param list_member_interval_ms: In HA mode, this client will request the
                                        living members periodically. A member means a server node
                                        of the Notification server cluster. This param specifies
                                        the interval of the listing member requests.
        :param retry_interval_ms: In HA mode a rpc call has failed on all the
                                  living members, this client will retry until success or timeout.
                                  This param specifies the retry interval.

        :param retry_timeout_ms: In HA mode a rpc call has failed on all the
                                 living members, this client will retry until success or timeout.
                                 This param specifies the retry timeout.
        :param sender: The identifier of the client.
        :param properties: The properties of NotificationClient
        """
        self.conf = {} if not properties else properties
        max_receive_message_length = -1 if GRPC_MAX_RECEIVE_MESSAGE_LENGTH not in self.conf \
            else int(self.conf.get(GRPC_MAX_RECEIVE_MESSAGE_LENGTH))
        self.server_uri = server_uri
        channel = grpc.insecure_channel(target=self.server_uri,
                                        options=[('grpc.max_receive_message_length', max_receive_message_length)])
        self._default_namespace = default_namespace
        self.notification_stub = notification_service_pb2_grpc.NotificationServiceStub(channel)
        self.threads = {}  # type: Dict[Any, List[threading.Thread]]
        self.lock = threading.Lock()
        self.enable_ha = enable_ha
        self.list_member_interval_ms = list_member_interval_ms
        self.retry_interval_ms = retry_interval_ms
        self.retry_timeout_ms = retry_timeout_ms
        self._sender = sender
        self._enable_idempotence = True
        server_uri_list = self.server_uri.split(",")

        if ENABLE_IDEMPOTENCE_CONFIG in self.conf \
                and self.conf.get(ENABLE_IDEMPOTENCE_CONFIG).strip().lower() == 'false':
            self._enable_idempotence = False
        self._client_id = None if CLIENT_ID not in self.conf else int(self.conf.get(CLIENT_ID))
        self._initial_seq_num = None if INITIAL_SEQUENCE_NUMBER not in self.conf \
            else int(self.conf.get(INITIAL_SEQUENCE_NUMBER))

        if len(server_uri_list) > 1 or self.enable_ha:
            self.living_members = []
            self.current_uri = None
            last_error = None
            for server_uri in server_uri_list:
                channel = grpc.insecure_channel(server_uri)
                notification_stub = notification_service_pb2_grpc.NotificationServiceStub(channel)
                try:
                    request = ListMembersRequest(timeout_seconds=0)
                    response = notification_stub.listMembers(request)
                    if response.return_code == ReturnStatus.SUCCESS:
                        self.living_members = [proto_to_member(proto).server_uri
                                               for proto in response.members]
                    else:
                        raise Exception(response.return_msg)
                    self.current_uri = server_uri
                    self.notification_stub = notification_stub
                    break
                except grpc.RpcError as e:
                    last_error = e
            if self.current_uri is None:
                raise Exception("No available server uri!") from last_error
            self.ha_change_lock = threading.Lock()
            self.ha_running = True
            self.notification_stub = self._wrap_rpcs(self.notification_stub, server_uri)
            self.list_member_thread = threading.Thread(target=self._list_members, daemon=True)
            self.list_member_thread.start()
        if self._enable_idempotence:
            if self._client_id is None:
                request = RegisterClientRequest(
                    client_meta=ClientMeta(namespace=self._default_namespace, sender=self._sender))
                response = self.notification_stub.registerClient(request)
                if response.return_code == ReturnStatus.SUCCESS:
                    self._client_id = response.client_id
                else:
                    raise Exception(response.return_msg)
            else:
                request = ClientIdRequest(client_id=self._client_id)
                response = self.notification_stub.isClientExists(request)
                if response.return_code != ReturnStatus.SUCCESS:
                    raise Exception("Failed to close notification client: {}".format(self))
                elif not response.is_exists:
                    raise Exception("Init notification client with a client id which have not registered.")
            seq_num = 0 if self._initial_seq_num is None else int(self._initial_seq_num)
            self.sequence_num_manager = SequenceNumberManager(seq_num)

    def close(self):
        if self._client_id is not None:
            request = ClientIdRequest(client_id=self._client_id)
            response = self.notification_stub.deleteClient(request)
            if response.return_code != ReturnStatus.SUCCESS:
                raise Exception("Failed to close notification client: {}".format(self))
        logging.info("The notification client:{} has been closed.".format(self))

    @property
    def client_id(self):
        """Int. Id of the Client."""
        return self._client_id

    @property
    def sequence_number(self):
        return self.sequence_num_manager.get_sequence_number()

    def send_event(self, event: BaseEvent):
        """
        Send event to Notification Service.

        :param event: the event updated.
        :return: The created event which has version and create time.
        """
        signature = str(uuid.uuid4())
        if self._enable_idempotence:
            seq_num = self.sequence_num_manager.get_sequence_number()
            signature = '_'.join(['client', str(self._client_id), str(seq_num)])

        request = SendEventRequest(
            event=EventProto(
                key=event.key,
                value=event.value,
                event_type=event.event_type,
                context=event.context,
                namespace=self._default_namespace,
                sender=self._sender
            ),
            uuid=signature,
            enable_idempotence=self._enable_idempotence)
        response = self.notification_stub.sendEvent(request)
        if response.return_code == ReturnStatus.SUCCESS:
            if self._enable_idempotence:
                self.sequence_num_manager.increment_sequence_number()
            return event_proto_to_event(response.event)
        else:
            raise Exception(response.return_msg)

    def list_events(self,
                    key: Union[str, List[str]],
                    namespace: str = None,
                    version: int = None,
                    event_type: str = None,
                    start_time: int = None,
                    sender: str = None) -> List[BaseEvent]:
        """
        List specific events in Notification Service.

        :param key: Key of the event for listening.
        :param namespace: (Optional) Namespace of the event for listening.
        :param version: (Optional) Version of the events must greater than this version.
        :param event_type: (Optional) Type of the events.
        :param start_time: (Optional) Start time of the events.
        :param sender: The event sender.
        :return: The event list.
        """
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        request = ListEventsRequest(
            keys=key,
            start_version=version,
            event_type=event_type,
            start_time=start_time,
            namespace=self._default_namespace if namespace is None else namespace,
            sender=sender
        )
        response = self.notification_stub.listEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            if response.events is None:
                return []
            else:
                events = []
                for event_proto in response.events:
                    event = event_proto_to_event(event_proto)
                    events.append(event)
                return events
        else:
            raise Exception(response.return_msg)

    def count_events(self,
                     key: Union[str, List[str]],
                     namespace: str = None,
                     version: int = None,
                     event_type: str = None,
                     start_time: int = None,
                     sender: str = None) -> Tuple[int, List[SenderEventCount]]:
        """
        Count specific events in Notification Service.

        :param key: The key or the list of keys of the events for listening.
        :param namespace: (Optional) Namespace of the event for listening.
        :param version: (Optional) Version of the events must greater than this version.
        :param event_type: (Optional) Type of the events.
        :param start_time: (Optional) Start time of the events.
        :param sender: The event sender.
        :return: The total event count and the list of event counts of each sender.
        """
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        request = CountEventsRequest(
            keys=key,
            start_version=version,
            event_type=event_type,
            start_time=start_time,
            namespace=self._default_namespace if namespace is None else namespace,
            sender=sender
        )
        response = self.notification_stub.countEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            sender_event_counts = []
            for sender_event_count_proto in response.sender_event_counts:
                sender_event_count = event_count_proto_to_event_count(sender_event_count_proto)
                sender_event_counts.append(sender_event_count)
            return response.event_count, sender_event_counts
        else:
            raise Exception(response.return_msg)

    def start_listen_event(self,
                           key: Union[str, Tuple[str]],
                           watcher: EventWatcher,
                           namespace: str = None,
                           version: int = None,
                           event_type: str = None,
                           start_time: int = None,
                           sender: str = None) -> EventWatcherHandle:
        """
        Start listen specific `key` or `version` notifications in Notification Service.

        :param key: Key of notification for listening.
        :param watcher: Watcher instance for listening.
        :param namespace: (Optional) Namespace of the event for listening.
        :param version: (Optional) The version of the events must greater than this version.
        :param event_type: (Optional) Type of the events for listening.
        :param start_time: (Optional) Start time of the events for listening.
        :param sender: The event sender.
        :return: The handle used to stop the listening.
        """
        if isinstance(key, str):
            key = (key,)
        elif isinstance(key, Iterable):
            key = tuple(key)
        namespace = self._default_namespace if namespace is None else namespace
        sender = ANY_CONDITION if sender is None else sender
        event_type = ANY_CONDITION if event_type is None else event_type

        def list_events(client,
                        k: Tuple[str],
                        v: List[int],
                        t: str = None,
                        ts: int = None,
                        ns: str = None,
                        sd: str = None,
                        timeout_seconds: int = None):
            request = ListEventsRequest(
                keys=k,
                event_type=t,
                start_time=ts,
                start_version=v,
                timeout_seconds=timeout_seconds,
                namespace=ns,
                sender=sd
            )
            response = client.notification_stub.listEvents(request)
            if response.return_code == ReturnStatus.SUCCESS:
                if response.events is None:
                    return None
                else:
                    events = []
                    for event_proto in response.events:
                        event = event_proto_to_event(event_proto)
                        events.append(event)
                    return events
            else:
                raise Exception(response.return_msg)

        def listen(client, k, v, t, ts, ns, sd, w):
            th = threading.current_thread()
            listen_version = v
            while getattr(th, '_flag', True):
                notifications = list_events(
                    client,
                    k,
                    listen_version,
                    t,
                    ts,
                    ns,
                    sd,
                    NOTIFICATION_TIMEOUT_SECONDS)
                if len(notifications) > 0:
                    w.process(notifications)
                    listen_version = notifications[len(notifications) - 1].version

        thread = threading.Thread(
            target=listen,
            args=(self, key, version, event_type, start_time, namespace, sender,
                  watcher),
            daemon=True)
        thread.start()
        self.lock.acquire()
        try:
            if self.threads.get((key, namespace, event_type, sender)) is None:
                self.threads[(key, namespace, event_type, sender)] = []
            self.threads[(key, namespace, event_type, sender)].append(thread)
        finally:
            self.lock.release()
        return ThreadEventWatcherHandle(thread, (key, namespace, event_type, sender), self)

    def stop_listen_event(self, key: Union[str, Tuple[str]] = None,
                          namespace: str = None,
                          event_type: str = None,
                          sender: str = None):
        """
        Stop listen specific `key` notifications in Notification Service.

        :param key: Keys of notification for listening.
        :param namespace: (Optional) Namespace of notification for listening.
        :param event_type: (Optional) Type of the events for listening.
        :param sender: The event sender.
        """
        namespace = self._default_namespace if namespace is None else namespace
        event_type = ANY_CONDITION if event_type is None else event_type
        sender = ANY_CONDITION if sender is None else sender
        if key is None:
            for (thread_key, v) in self.threads.items():
                if thread_key == ALL_EVENTS_KEY:
                    # do not stop the global listen threads,
                    # which are controlled by `stop_listen_events`.
                    continue
                threads = self.threads[thread_key]
                for thread in threads:
                    thread._flag = False
                    thread.join()
            self.threads.clear()
        else:
            self.lock.acquire()
            if isinstance(key, str):
                key = (key,)
            try:
                if (key, namespace, event_type, sender) in self.threads:
                    threads = self.threads[(key, namespace, event_type, sender)]
                    for thread in threads:
                        thread._flag = False
                        thread.join()
                    del self.threads[(key, namespace, event_type, sender)]
            finally:
                self.lock.release()

    def list_all_events(self,
                        start_time: int = None,
                        start_version: int = None,
                        end_version: int = None) -> List[BaseEvent]:
        """
        List specific `key` or `version` of events in Notification Service.

        :param start_time: (Optional) Start time of the events.
        :param start_version: (Optional) the version of the events must greater than the
                              start_version.
        :param end_version: (Optional) the version of the events must equal or less than the
                            end_version.
        :return: The event list.
        """
        request = ListAllEventsRequest(start_time=start_time,
                                       start_version=start_version,
                                       end_version=end_version)
        response = self.notification_stub.listAllEvents(request)
        if response.return_code == ReturnStatus.SUCCESS:
            if response.events is None:
                return []
            else:
                events = []
                for event_proto in response.events:
                    event = event_proto_to_event(event_proto)
                    events.append(event)
                return events
        else:
            raise Exception(response.return_msg)

    def start_listen_events(self,
                            watcher: EventWatcher,
                            start_time=int(time.time() * 1000),
                            version: int = None) -> EventWatcherHandle:
        """
        Start listen all events.

        :param watcher: process event.
        :param start_time: (Optional) the earliest event time.
        :param version: (Optional) the start version of the event.
        :return: The handle used to stop the listening.
        """
        if ALL_EVENTS_KEY in self.threads:
            raise Exception("already listen events, must stop first!")

        def list_events(client, start, timeout_seconds: int = None):
            request = ListAllEventsRequest(start_time=start, timeout_seconds=timeout_seconds)
            response = client.notification_stub.listAllEvents(request)
            if response.return_code == ReturnStatus.SUCCESS:
                if response.events is None:
                    return None
                else:
                    events = []
                    for event_proto in response.events:
                        event = event_proto_to_event(event_proto)
                        events.append(event)
                    return events
            else:
                raise Exception(response.return_msg)

        def list_events_from_version(client, v, timeout_seconds: int = None):
            request = ListAllEventsRequest(start_version=v, timeout_seconds=timeout_seconds)
            response = client.notification_stub.listAllEvents(request)
            if response.return_code == ReturnStatus.SUCCESS:
                if response.events is None:
                    return None
                else:
                    events = []
                    for event_proto in response.events:
                        event = event_proto_to_event(event_proto)
                        events.append(event)
                    return events
            else:
                raise Exception(response.return_msg)

        def listen(client, s, v, w):
            t = threading.current_thread()
            flag = True if v is None else False
            current_version = 0 if v is None else v
            try:
                while getattr(t, '_flag', True):
                    if flag:
                        notifications = list_events(client, s, NOTIFICATION_TIMEOUT_SECONDS)
                        if len(notifications) > 0:
                            w.process(notifications)
                            current_version = notifications[len(notifications) - 1].version
                            flag = False
                    else:
                        notifications = list_events_from_version(client,
                                                                 current_version,
                                                                 NOTIFICATION_TIMEOUT_SECONDS)
                        if len(notifications) > 0:
                            w.process(notifications)
                            current_version = notifications[len(notifications) - 1].version
            except Exception as e:
                logging.exception("Exception when listening events, %s", e)
                raise e

        thread = threading.Thread(target=listen,
                                  args=(self, start_time, version, watcher),
                                  daemon=True)
        thread.start()
        self.lock.acquire()
        try:
            if self.threads.get(ALL_EVENTS_KEY) is None:
                self.threads[ALL_EVENTS_KEY] = []
            self.threads[ALL_EVENTS_KEY].append(thread)
        finally:
            self.lock.release()
        return ThreadEventWatcherHandle(thread, ALL_EVENTS_KEY, self)

    def stop_listen_events(self):
        """
        Stop the global listening threads.
        """
        self.lock.acquire()
        try:
            if ALL_EVENTS_KEY in self.threads:
                threads = self.threads[ALL_EVENTS_KEY]
                for thread in threads:
                    thread._flag = False
                    thread.join()
                del self.threads[ALL_EVENTS_KEY]
        finally:
            self.lock.release()

    def get_latest_version(self, key: str = None, namespace: str = None):
        """
        get latest event's version by key.
        :param key: (Optional) Key of notification for listening.
        :param namespace: (Optional) Namespace of notification for listening.
        :return: Version number of the specific key.
        """
        self.lock.acquire()
        try:
            request = GetLatestVersionByKeyRequest(key=key,
                                                   namespace=self._default_namespace if namespace is None else namespace)
            response = self.notification_stub.getLatestVersionByKey(request)
            if response.return_code == str(ReturnStatus.SUCCESS):
                return response.version
        finally:
            self.lock.release()

    def disable_high_availability(self):
        if hasattr(self, "ha_running"):
            self.ha_running = False
            self.list_member_thread.join()

    def _list_members(self):
        while self.ha_running:
            # refresh the living members
            request = ListMembersRequest(timeout_seconds=int(self.list_member_interval_ms / 1000))
            response = self.notification_stub.listMembers(request)
            if response.return_code == ReturnStatus.SUCCESS:
                with self.ha_change_lock:
                    self.living_members = [proto_to_member(proto).server_uri
                                           for proto in response.members]
            else:
                logging.error("Exception thrown when updating the living members: %s" %
                              response.return_msg)
            time.sleep(int(self.list_member_interval_ms / 1000))

    def _ha_wrapper(self, func):
        @wraps(func)
        def call_with_retry(*args, **kwargs):
            current_func = getattr(self.notification_stub,
                                   func.__name__).inner_func
            start_time = time.time_ns() / 1000000
            failed_members = set()
            while True:
                try:
                    return current_func(*args, **kwargs)
                except grpc.RpcError:
                    logging.error("Exception thrown when calling rpc, change the connection.",
                                  exc_info=True)
                    with self.ha_change_lock:
                        # check the current_uri to ensure thread safety
                        if current_func.server_uri == self.current_uri:
                            living_members = list(self.living_members)
                            failed_members.add(self.current_uri)
                            shuffle(living_members)
                            found_new_member = False
                            for server_uri in living_members:
                                if server_uri in failed_members:
                                    continue
                                next_uri = server_uri
                                channel = grpc.insecure_channel(next_uri)
                                notification_stub = self._wrap_rpcs(
                                    notification_service_pb2_grpc.NotificationServiceStub(channel),
                                    next_uri)
                                self.notification_stub = notification_stub
                                current_func = getattr(self.notification_stub,
                                                       current_func.__name__).inner_func
                                self.current_uri = next_uri
                                found_new_member = True
                            if not found_new_member:
                                logging.error("No available living members currently. Sleep and retry.")
                                failed_members.clear()
                                sleep_and_detecting_running(self.retry_interval_ms,
                                                            lambda: self.ha_running)

                # break if stopped or timeout
                if not self.ha_running or \
                        time.time_ns() / 1000000 > start_time + self.retry_timeout_ms:
                    if not self.ha_running:
                        raise Exception("HA has been disabled.")
                    else:
                        raise Exception("Rpc retry timeout!")

        call_with_retry.inner_func = func
        return call_with_retry

    def _wrap_rpcs(self, stub, server_uri):
        for method_name, method in dict(stub.__dict__).items():
            method.__name__ = method_name
            method.server_uri = server_uri
            setattr(stub, method_name, self._ha_wrapper(method))
        return stub

    def __str__(self):
        return "NotificationClient(id={}, namespace={}, sender={}, server_uri={}, properties={})".format(
            self._client_id,
            self._default_namespace,
            self._sender,
            self.server_uri,
            self.conf
        )



#!/usr/bin/env python
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import argparse
import os
import logging
from typing import Dict
from notification_service.util import db
from notification_service.server_config import NotificationServerConfig
from notification_service.server import NotificationServerRunner


def create_server_config(root_dir_path, param: Dict[str, str]):
    """
    Generate default server config which use Apache Airflow as scheduler.
    """
    import notification_service.config_templates
    default_config_file_name = 'default_notification_server.yaml'
    config_file_name = 'notification_server.yaml'

    if not os.path.exists(root_dir_path):
        logging.info("{} does not exist, creating the directory".format(root_dir_path))
        os.makedirs(root_dir_path, exist_ok=False)
    default_server_config_path = os.path.join(
        os.path.dirname(notification_service.config_templates.__file__), default_config_file_name
    )
    if not os.path.exists(default_server_config_path):
        raise Exception("default notification server config is not found at {}.".format(default_server_config_path))

    server_config_target_path = os.path.join(root_dir_path, config_file_name)
    with open(default_server_config_path, encoding='utf-8') as config_file:
        default_config = config_file.read().format(**param)
    with open(server_config_target_path, mode='w', encoding='utf-8') as f:
        f.write(default_config)
    logging.info("Notification server config generated at {}".format(server_config_target_path))
    return server_config_target_path


def create_all_tables(db_uri):
    db.create_all_tables(db_uri)


def drop_all_tables(db_uri):
    db.drop_all_tables(db_uri)


def _prepare_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-config-only', default=None, action='store_true',
                        help='Only generate notification server configuration file.')
    return parser.parse_args()


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s',
                        level=logging.INFO)
    if "NOTIFICATION_HOME" not in os.environ:
        os.environ["NOTIFICATION_HOME"] = os.environ["HOME"] + "/notification_service"
        logging.info("Set env variable NOTIFICATION_HOME to {}".format(os.environ["NOTIFICATION_HOME"]))
    if not os.path.exists(os.environ["NOTIFICATION_HOME"]):
        os.makedirs(os.environ["NOTIFICATION_HOME"])
    config_file = os.environ["NOTIFICATION_HOME"] + "/notification_server.yaml"

    args = _prepare_args()

    if not os.path.exists(config_file):
        create_server_config(os.environ.get("NOTIFICATION_HOME"), os.environ.copy())
    else:
        logging.info("Notification server config exists at {}".format(config_file))

    if args.generate_config_only:
        exit(0)

    config = NotificationServerConfig(config_file)
    create_all_tables(db_uri=config.db_uri)
    ns = NotificationServerRunner(config_file=config_file)
    logging.info('notification server start(port:{}).'.format(ns.config.port))
    ns.start(is_block=True)
